"""Helper classes and methods to access and process Landsat Level-1 data.

Glossary of variables referring to a band with examples :

  * file name : 'LE07_L1TP_195049_20000422_20170212_01_T1_B4.TIF'
  * file suffix : 'B4'
  * band number : 4
  * band long name : 'Near Infrared (NIR)'
  * band short name : 'nir'
"""

from itertools import chain
from datetime import datetime
import json
import os
from pkg_resources import resource_string

import numpy as np
import rasterio
from rasterio.path import parse_path

from pylandsat import preprocessing


BANDS = json.loads(resource_string(__name__, 'bands.json'))


def _to_numeric(value):
    """Try to convert a string to an integer or a float.
    If not possible, returns the initial string.
    """
    try:
        value  = int(value)
    except:
        try:
            value = float(value)
        except:
            pass
    return value


def _suffix_from_fname(file_name):
    """Get file suffix from file name."""
    if os.path.sep in file_name:
        file_name = os.path.basename(file_name)
    return file_name[41:].split('.')[0]


def _suffix_from_bnum(band_number):
    """Get file suffix from band number."""
    if band_number in (61, 62):
        return 'B' + str(band_number)[0] + '_VCID_' + str(band_number)[1]
    return 'B' + str(band_number)


def _suffix_from_name(band_name, sensor):
    """Get file suffix from band name (either long or short)."""
    if sensor not in BANDS:
        raise ValueError('Sensor %s not found.' % sensor)
    for suffix, long_name in BANDS[sensor].items():
        if band_name in (long_name, _band_shortname(long_name)):
            return suffix
    raise ValueError('Band %s not found.' % band_name)


def _is_band(suffix):
    """Determine if a file suffix refers to a band."""
    return suffix[0] == 'B' and suffix[1].isnumeric()


def _band_number(suffix):
    """Get band number from file suffix."""
    if _is_band(suffix):
        bnum_str = suffix[1:]
        # B6_VCID_1 = 61 & B6_VCID_2 = 62
        if 'VCID' in bnum_str:
            bnum = int(bnum_str[0] + bnum_str[-1])
        else:
            bnum = int(bnum_str)
    else:
        raise ValueError('The provided suffix does not refer to a band.')
    return bnum


def _band_shortname(long_name):
    """Get short band name, e.g. `Near Infrared (NIR)` becomes `nir` and
    `Red` becomes `red`.
    """
    if '(' in long_name:
        start = long_name.find('(') + 1
        end = long_name.find(')')
        short_name = long_name[start:end]
    else:
        short_name = long_name.replace(' ', '_').replace('-', '_')
    return short_name.lower()


class Scene:
    def __init__(self, scene_dir):
        """Landsat Level-1 scene."""
        self.dir = os.path.abspath(scene_dir)
        self.mtl = self._parse_mtl()

    def __getattr__(self, name):
        """Returns a Band object if possible."""
        if name in self.available_bands():
            return Band(self, _suffix_from_name(name, self.sensor))
        raise AttributeError()
    
    def __iter__(self):
        """Iterate over bands."""
        available = self.available_bands()
        for suffix, long_name in BANDS[self.sensor].items():
            short_name = _band_shortname(long_name)
            if short_name in available:
                yield Band(self, suffix)

    def _available_files(self):
        """List available files in the scene directory."""
        return [f for f in os.listdir(self.dir)
                if f.endswith('.TIF') or f.endswith('.txt')]

    def available_bands(self):
        """List short names of available bands."""
        bands = []
        for fname in self._available_files():
            suffix = _suffix_from_fname(fname)
            if _is_band(suffix):
                long_name = BANDS[self.sensor][suffix]
                bands.append(_band_shortname(long_name))
        return bands

    def file_path(self, suffix):
        """Find file path according to a given file suffix."""
        for fname in self._available_files():
            if suffix == _suffix_from_fname(fname):
                return os.path.join(self.dir, fname)
        raise ValueError('Suffix %s not available.' % suffix)

    def _parse_mtl(self):
        """Parse MTL metadata file.

        Returns
        -------
        mtl : dict
            MTL metadata in a dictionnary.
        """
        mtl = {}
        fpath = self.file_path('MTL')
        f = open(fpath)
        for line in f.readlines():
            if line.startswith('GROUP = L1_METADATA_FILE'):
                pass  # ignore main group
            elif line.strip().startswith('END'):
                pass  # ignore end statements
            elif line.strip().startswith('GROUP'):
                group = line.split('=')[-1].strip()
                mtl[group] = {}
            else:
                param, value = [s.strip() for s in line.split('=')]
                value = value.replace('"', '')
                value = _to_numeric(value)
                mtl[group][param] = value
        return mtl

    @property
    def scene_id(self):
        """Landsat scene identifier."""
        if 'SCENE_ID' in self.mtl['METADATA_FILE_INFO']:
            return self.mtl['METADATA_FILE_INFO']['SCENE_ID']
        else:
            return self.mtl['METADATA_FILE_INFO']['LANDSAT_SCENE_ID']

    @property
    def product_id(self):
        """Landsat product identifier."""
        if 'PRODUCT_ID' in self.mtl['METADATA_FILE_INFO']:
            return self.mtl['METADATA_FILE_INFO']['PRODUCT_ID']
        else:
            return self.mtl['METADATA_FILE_INFO']['LANDSAT_PRODUCT_ID']

    @property
    def spacecraft(self):
        """Spacecraft ID."""
        return self.mtl['PRODUCT_METADATA']['SPACECRAFT_ID']

    @property
    def sensor(self):
        """Sensor ID."""
        sensor = self.mtl['PRODUCT_METADATA']['SENSOR_ID']
        if sensor == 'MSS' and self.spacecraft in ('LANDSAT_4', 'LANDSAT_5'):
            sensor += '_'
        return sensor

    @property
    def date(self):
        """Acquisition date."""
        date_acquired = self.mtl['PRODUCT_METADATA']['DATE_ACQUIRED']
        return datetime.strptime(date_acquired, '%Y-%m-%d')

    @property
    def wrs_path(self):
        """WRS2 Path."""
        return self.mtl['PRODUCT_METADATA']['WRS_PATH']

    @property
    def wrs_row(self):
        """WRS2 Row."""
        return self.mtl['PRODUCT_METADATA']['WRS_ROW']
    
    @property
    def quality(self):
        """Quality band."""
        return Band(self, 'BQA')


class Band(rasterio.io.DatasetReader):

    def __init__(self, scene, suffix):
        self.scene = scene
        self.suffix = suffix
        self.fname = '{pid}_{suffix}.TIF'.format(
            pid=scene.product_id,
            suffix=suffix
        )
        self.fpath = os.path.join(scene.dir, self.fname)
        if suffix == 'BQA':
            self.long_name = 'Quality Band'
            self.bname = 'bqa'
            self.bnum = None
        else:
            self.long_name = BANDS[self.scene.sensor][suffix]
            self.bname = _band_shortname(self.long_name)
            self.bnum = _band_number(self.suffix)
        super().__init__(parse_path(self.fpath))

    def _gain_bias(self, unit='radiance'):
        """Get band-specific radiometric rescaling factor.
        Returns a (gain, bias) tuple. Unit can be radiance or
        reflectance.
        """
        if self.bnum == 61:
            bnum = '6_VCID_1'
        elif self.bnum == 62:
            bnum = '6_VCID_2'
        else:
            bnum = str(self.bnum)
        key_gain = unit.upper() + '_MULT_BAND_' + bnum
        key_bias = unit.upper() + '_ADD_BAND_' + bnum
        gain = self.scene.mtl['RADIOMETRIC_RESCALING'][key_gain]
        bias = self.scene.mtl['RADIOMETRIC_RESCALING'][key_bias]
        return gain, bias

    def _k1_k2(self):
        """Get band-specific thermal constants."""
        if self.bnum == 61:
            bnum = '6_VCID_1'
        elif self.bnum == 62:
            bnum = '6_VCID_2'
        else:
            bnum = str(self.bnum)
        k1_key = 'K1_CONSTANT_BAND_' + bnum
        k2_key = 'K2_CONSTANT_BAND_' + bnum
        if self.scene.spacecraft == 'LANDSAT_8':
            k1 = self.scene.mtl['TIRS_THERMAL_CONSTANTS'][k1_key]
            k2 = self.scene.mtl['TIRS_THERMAL_CONSTANTS'][k2_key]
        else:
            k1 = self.scene.mtl['THERMAL_CONSTANTS'][k1_key]
            k2 = self.scene.mtl['THERMAL_CONSTANTS'][k2_key]
        return k1, k2

    def to_radiance(self, custom_dn=None):
        """Convert DN values to TOA spectral radiance."""
        if isinstance(custom_dn, np.ndarray):
            dn = custom_dn
        else:
            dn = self.read(1)
        gain, bias = self._gain_bias(unit='radiance')
        return preprocessing.to_radiance(dn, gain, bias)

    def to_reflectance(self, custom_dn=None):
        """Convert DN values to TOA spectral reflectance."""
        if 'tir' in self.bname:
            raise ValueError('TIR bands cannot be converted to reflectance.')
        if isinstance(custom_dn, np.ndarray):
            dn = custom_dn
        else:
            dn = self.read(1)
        gain, bias = self._gain_bias(unit='reflectance')
        return preprocessing.to_reflectance(dn, gain, bias)

    def to_brightness_temperature(self, custom_dn=None):
        """Convert DN values to brightness temperature."""
        if not 'tir' in self.bname:
            raise ValueError('Only thermal bands can be converted '
                             'to brightness temperature.')
        if isinstance(custom_dn, np.ndarray):
            dn = custom_dn
        else:
            dn = self.read(1)
        radiance = self.to_radiance(custom_dn=dn)
        k1, k2 = self._k1_k2()
        return preprocessing.to_brightness_temperature(radiance, k1, k2)



"""Downloading Landsat products from Google public dataset."""

import json
import os
from pkg_resources import resource_string

from pylandsat.utils import download_file, meta_from_pid

BASE_URL = ('https://storage.googleapis.com/gcp-public-data-landsat/'
            '{sensor}/{collection:02}/{path:03}/{row:03}/{product_id}/')


class Product:
    """Landsat product to download."""
    def __init__(self, product_id):
        """Initialize a product download.

        Attributes
        ----------
        product_id : str
            Landsat product identifier.
        out_dir : str
            Path to output directory.
        """
        self.product_id = product_id
        self.meta = meta_from_pid(product_id)
        self.baseurl = BASE_URL.format(**self.meta)

    @property
    def available(self):
        """List all available files."""
        resource = resource_string(__name__, 'files.json')
        labels = json.loads(resource)
        return labels[self.meta['sensor']]

    def _url(self, label):
        """Get download URL of a given file according to its label."""
        if 'README' in label:
            basename = label
        else:
            basename = self.product_id + '_' + label
        return self.baseurl + basename

    def download(self, out_dir, progressbar=True, files=None):
        """Download a Landsat product.

        Parameters
        ----------
        out_dir : str
            Path to output directory. A subdirectory named after the
            product ID will automatically be created.
        progressbar : bool, optional
            Show a progress bar.
        files : list of str, optional
            Specify the files to download manually. By default, all available
            files will be downloaded.
        """
        dst_dir = os.path.join(out_dir, self.product_id)
        os.makedirs(dst_dir, exist_ok=True)
        if not files:
            files = self.available
        else:
            files = [f for f in files if f in self.available]

        for label in files:
            if '.tif' in label:
                label = label.replace('.tif', '.TIF')
            url = self._url(label)
            download_file(url, dst_dir, progressbar=progressbar, verify=True)



import gzip
import os
import hashlib
from datetime import datetime

import rasterio
import requests
from tqdm import tqdm


def bounds_from_transform(transform, width, height):
    """Calculate raster bounds from transform, width & height."""
    xres, yres = transform.a, transform.e
    left, top = transform.c, transform.f
    right = left + xres * width
    bottom = top + yres * height
    return rasterio.coords.BoundingBox(left, bottom, right, top)


def meta_from_pid(product_id):
    """Extract metadata contained in a Landsat Product Identifier."""
    meta = {}
    parts = product_id.split('_')
    meta['product_id'] = product_id
    meta['sensor'], meta['correction'] = parts[0], parts[1]
    meta['path'], meta['row'] = int(parts[2][:3]), int(parts[2][3:])
    meta['acquisition_date'] = datetime.strptime(parts[3], '%Y%m%d')
    meta['processing_date'] = datetime.strptime(parts[4], '%Y%m%d')
    meta['collection'], meta['tier'] = int(parts[5]), parts[6]
    return meta


def compute_md5(fpath):
    """Get hexadecimal MD5 hash of a file."""
    with open(fpath, 'rb') as f:
        h = hashlib.md5(f.read())
    return h.hexdigest()


def _size(headers):
    """Get size of a remote file in bytes by parsing the HTTP headers.
    Returns 0 if the information is not available.
    """
    if 'x-goog-stored-content-length' in headers:
        length = headers['x-goog-stored-content-length']
    elif 'Content-Length' in headers:
        length = headers['Content-Length']
    else:
        length = 0
    return int(length)


def download_file(url, outdir, progressbar=False, verify=False):
    """Download a file from an URL into a given directory.

    Parameters
    ----------
    url : str
        File to download.
    outdir : str
        Path to output directory.
    progressbar : bool, optional
        Display a progress bar.
    verify : bool, optional
        Check that remote and local MD5 haches are equal.
    
    Returns
    -------
    fpath : str
        Path to downloaded file.
    """
    fname = url.split('/')[-1]
    fpath = os.path.join(outdir, fname)
    r = requests.get(url, stream=True)
    remotesize = int(r.headers.get('Content-Length', 0))
    etag = r.headers.get('ETag', '').replace('"', '')

    if r.status_code != 200:
        raise requests.exceptions.HTTPError(str(r.status_code))

    if os.path.isfile(fpath) and os.path.getsize(fpath) == remotesize:
        return fpath
    if progressbar:
        progress = tqdm(total=remotesize, unit='B', unit_scale=True)
        progress.set_description(fname)
    with open(fpath, 'wb') as f:
        for chunk in r.iter_content(chunk_size=1024 * 1024):
            if chunk:
                f.write(chunk)
                if progressbar:
                    progress.update(1024 * 1024)

    r.close()
    if progressbar:
        progress.close()

    if verify:
        if not compute_md5(fpath) == etag:
            raise requests.exceptions.HTTPError('Download corrupted.')

    return fpath


def decompress(fpath, remove_archive=False):
    """Decompress a .gz archive. Returns path to decompressed file."""
    basedir, fname = os.path.dirname(fpath), os.path.basename(fpath)
    outpath = os.path.join(basedir, fname.replace('.gz', ''))
    with gzip.open(fpath, 'rb') as src, open(outpath, 'wb') as dst:
        for line in src:
            dst.write(line)
    if remove_archive:
        os.remove(fpath)
    return outpath



"""Convert quantized and calibrated scaled Digital Numbers (DN) to
top-of-atmosphere (TOA) reflectance, radiance or brightness temperature.
"""

import numpy as np


def to_radiance(dn, gain, bias):
    """Convert DN to TOA spectral radiance.

    Parameters
    ----------
    dn : 2d array
        Input image, DN values.
    gain : float
        Band-specific multiplicative rescaling factor.
    bias : float
        Band-specific additive rescaling factor.
    
    Returns
    -------
    radiance : 2d array
        Output image, TOA spectral radiance.
    """
    return gain * dn + bias


def to_reflectance(dn, gain, bias, sun_elevation_angle=None):
    """Convert DN to TOA spectral reflectance with an optional correction
    for the sun angle.

    Parameters
    ----------
    dn : 2d array
        Input image, DN values.
    gain : float
        Band-specific multiplicative rescaling factor.
    bias : float
        Band-specific additive rescaling factor.
    sun_elevation_angle : float, optional
        Sun elevation angle in degrees.
    
    Returns
    -------
    reflectance : 2d array
        Output image, TOA planetary reflectance.
    """
    reflectance = gain * dn + bias
    if sun_elevation_angle:
        reflectance /= np.sin(np.radians(sun_elevation_angle))
    return reflectance


def to_brightness_temperature(radiance, k1, k2):
    """Convert TOA spectral radiance to TOA brightness temperature.

    Parameters
    ----------
    radiance : 2d array
        Input image, TOA spectral radiance.
    k1 : float
        Band-specific thermal conversion constant.
    k2 : float
        Band-specific thermal conversion constant.
    
    Returns
    -------
    bt : 2d array
        Output image, TOA brightness temperature.
    """
    return k2 / np.log(k1 / (radiance + 1))


"""Command-line interface."""

import csv
import json
import os
from pkg_resources import resource_string

from appdirs import user_data_dir
import click
from shapely.geometry import shape, Point

try:
    from geopy import Nominatim
    _has_geopy = True
except ImportError:
    _has_geopy = False
    pass

from pylandsat.catalog import Catalog
from pylandsat.database import sync_catalog, sync_wrs
from pylandsat.download import Product


@click.group()
def cli():
    pass


@click.command(name='print-data-dir')
def print_datadir():
    """Print user data directory where database is stored."""
    click.echo(user_data_dir(appname='pylandsat'))


@click.command(name='list-sensors')
def list_sensors():
    """Print supported sensors."""
    res = resource_string(__name__, 'files.json')
    files = json.loads(res)
    click.echo(', '.join(files.keys()))


@click.command(name='list-available-files')
@click.argument('sensor', type=click.STRING)
def list_available_files(sensor):
    """Print available sensors."""
    res = resource_string(__name__, 'files.json')
    files = json.loads(res)
    if sensor not in files:
        click.exceptions.BadArgumentUsage('Sensor not supported.')
    click.echo(', '.join(files[sensor]))


@click.command(name='sync-database')
@click.option('-f', '--force', default=False, is_flag=True,
              help='Overwrite existing database.')
def syncdb(force):
    """Download Landsat Catalog and WRS2 shapefile and export the data to a
    local SQLite database.
    """
    data_dir = user_data_dir(appname='pylandsat')
    dbpath = os.path.join(data_dir, 'landsat.db')
    if os.path.isfile(dbpath) and not force:
        click.echo('Database already exists.')
    else:
        click.echo('Syncing catalog...')
        sync_catalog()
        click.echo('Syncing WRS2 paths/rows...')
        sync_wrs()


def _geom_from_geojson(fpath):
    """Get shapely geometry from a GeoJSON file. If multiple features
    are available, only the first one is used.
    """
    with open(fpath) as f:
        geojson = json.load(f)
    if geojson['type'] == 'Feature':
        geom = shape(geojson['geometry'])
    elif geojson['type'] == 'FeatureCollection':
        geom = shape(geojson['features'][0]['geometry'])
    else:
        raise click.exceptions.BadOptionUsage('No GeoJSON feature found.')
    return geom


def _geom_from_address(address):
    """Get a Shapely point geometry by geocoding an address with
    Nominatim.
    """
    if not _has_geopy:
        raise ImportError('Geopy is required for address geocoding.')
    geoloc = Nominatim(user_agent=__name__)
    loc = geoloc.geocode(address)
    return Point(loc.longitude, loc.latitude)

def _to_csv(records, output_file):
    """Write a list of dicts to a CSV file."""
    if not records:
        return
    with open(output_file, 'w') as dst:
        writer = csv.DictWriter(dst, records[0].keys())
        writer.writeheader()
        for record in records:
            if 'sensing_time' in record:
                date = record['sensing_time'].strftime('%Y-%m-%d')
                record.update(sensing_time=date)
            writer.writerow(record)


@click.command(name='search')
@click.option('-b', '--begin', type=click.STRING,
              help='Begin search date (YYYY-MM-DD).')
@click.option('-e', '--end', type=click.STRING,
              help='End search date (YYYY-MM-DD).')
@click.option('-g', '--geojson', type=click.Path(exists=True), default=None,
              help='Area of interest (GeoJSON file).')
@click.option('-l', '--latlon', nargs=2, type=click.FLOAT, default=None,
              help='Point of interest (decimal lat/lon).')
@click.option('-a', '--address', type=click.STRING, default=None,
              help='Address of interest.')
@click.option('-p', '--path', type=click.INT, default=None,
              help='WRS2 path.')
@click.option('-r', '--row', type=click.INT, default=None,
              help='WRS2 row.')
@click.option('-c', '--clouds', type=click.FLOAT, default=None,
              help='Max. cloud cover percentage.')
@click.option('-s', '--sensors', type=click.STRING, default=None,
              help='Comma-separated list of possible sensors.')
@click.option('-t', '--tiers', type=click.STRING, default=None,
              help='Comma-separated list of possible collection tiers.')
@click.option('--slcoff', is_flag=True,
              help='Include SLC-off LE7 scenes.')
@click.option('-o', '--output', default=None,
              type=click.Path(file_okay=True, writable=True),
              help='Output CSV file.')
def search(begin, end, geojson, latlon, address, path, row, clouds, sensors,
           tiers, slcoff, output):
    """Search for scenes in the Google Landsat Public Dataset catalog."""
    if geojson:
        geom = _geom_from_geojson(geojson)
    elif latlon:
        y, x = latlon
        geom = Point(x, y)
    elif address:
        geom = _geom_from_address(address)
    else:
        geom = None

    # If only year provided, set default month and day
    if len(begin) == 4:
        begin += '-01-01'
    if len(end) == 4:
        end += '-01-01'

    if sensors:
        sensors = [s.strip() for s in sensors.split(',')]
    if tiers:
        tiers = [t.trip() for t in tiers.split(',')]

    catalog = Catalog()
    scenes = catalog.search(begin, end, path, row, geom, clouds, sensors,
                            tiers, slc=not slcoff)

    if output:
        _to_csv(scenes, output)
    else:
        for scene in scenes:
            click.echo(scene['product_id'])


@click.command(name='download')
@click.argument('products', type=click.STRING, nargs=-1)
@click.option('-d', '--output-dir', type=click.Path(dir_okay=True), default='.',
              help='Output directory.')
@click.option('-f', '--files', type=click.STRING, default=None,
              help='Comma-separated list of files to download.')
def download(products, output_dir, files):
    """Download a Landsat product according to its identifier."""
    if files:
        files = [f.strip() for f in files.split(',')]
    for i, pid in enumerate(products):
        click.echo('Downloading {} ({}/{}).'.format(pid, i+1, len(products)))
        product = Product(pid)
        product.download(output_dir, progressbar=True, files=files)


cli.add_command(search)
cli.add_command(download)
cli.add_command(syncdb)
cli.add_command(print_datadir)
cli.add_command(list_sensors)
cli.add_command(list_available_files)


if __name__ == '__main__':
    cli()



"""Search for scenes in the Landsat catalog."""

from datetime import datetime

from shapely import wkt

from pylandsat import queries
from pylandsat.database import LandsatDB


SENSORS = ['LC08', 'LE07', 'LT05', 'LT04', 'LM05',
           'LM04', 'LM03', 'LM02', 'LM01']
TIERS = ['T1', 'T2', 'RT']


def _to_list(value):
    """Convert string parameter value to a list with a single element."""
    if not value:
        return None
    elif isinstance(value, list):
        return value
    elif isinstance(value, tuple):
        return list(value)
    else:
        return [value]


class Catalog:
    """Perform queries on the Landsat catalog."""

    def __init__(self):
        self.db = LandsatDB()

    def search(self, begin, end, path=None, row=None, geom=None,
               maxcloud=None, sensors=None, tiers=None, slc=True):
        """Search for scenes in the Landsat catalog.

        Parameters
        ----------
        begin : datetime
            Begin search period as a Python datetime or
            as a string (YYYY-MM-DD).
        end : datetime
            End search period as Python datetime or as
            a string (YYYY-MM-DD).
        path : int, optional
            WRS2 paths as an int or a list of int.
        row : int, optional
            WRS2 rows as an int or a list of int.
        geom : geometry, optional
            Area or point of interest as a shapely geometry.
        maxcloud : float, optional
            Max. cloud cover percentage.
        sensors : str, optional
            Sensors of interest as a str or a list of str.
        tiers : str, optional
            Collection tiers of interest as a str or a list of str.
        slc : bool, optional
            If true, exclude Landsat 7 products acquired after
            the SLC failure.

        Returns
        -------
        scenes : List
            Search results as a list.
        """
        # Convert strings to list if necessary
        path, row = _to_list(path), _to_list(row)
        sensors, tiers = _to_list(sensors), _to_list(tiers)

        # Convert str to datetimes and to integer timestamps
        if isinstance(begin, str):
            begin = datetime.strptime(begin, '%Y-%m-%d')
        if isinstance(end, str):
            end = datetime.strptime(end, '%Y-%m-%d')
        begin = int(begin.timestamp())
        end = int(end.timestamp())

        # Convert geometry to WKT
        if geom:
            geom = wkt.dumps(geom, rounding_precision=8)

        # Default values
        if not maxcloud and not isinstance(maxcloud, float):
            maxcloud = 100.
        if not sensors:
            sensors = SENSORS
        if not tiers:
            tiers = TIERS

        # Spatial filter with WRS paths and rows
        if path and row:
            scenes = self.db.query(
                query=queries.CATALOG_SEARCH_PATHROW,
                params=(path, row, begin, end, maxcloud, sensors, tiers))
        elif geom:
            scenes = self.db.query(
                query=queries.CATALOG_SEARCH_GEOM,
                params=(geom, begin, end, maxcloud, sensors, tiers, geom))
        else:
            raise ValueError('No spatial information provided.')
        
        for scene in scenes:
            scene.update(
                sensing_time=datetime.fromtimestamp(scene['sensing_time']))

        def _slc_on(scene):
            FAILURE = datetime(2003, 5, 31)
            sat = scene['product_id'][3]
            if sat != '7':
                return True
            elif scene['sensing_time'] >= FAILURE:
                return False
            else:
                return True
        
        if slc:
            scenes_ = scenes.copy()
            scenes = [scene for scene in scenes_ if _slc_on(scene)]

        return scenes

    def wrs(self, geom):
        """Find WRS2 paths and rows that intersect a given geometry.
        Also returns the geometry of each footprint and their coverage
        of the geometry of interest if it is not a polygon.
        """
        geom_wkt = wkt.dumps(geom, rounding_precision=8)
        path_row = self.db.query(
            queries.WRS_SEARCH, params=(geom_wkt, geom_wkt))

        if 'POLYGON' in geom_wkt:
            for pr in path_row:
                footprint = wkt.loads(pr['geom'])
                cover = geom.intersection(footprint).area / geom.area
                pr.update(cover=cover)

        return path_row

    def metadata(self, product_id):
        """Get available metadata for a given scene identified
        by its Product Identifier.
        """
        response = self.db.query(
            queries.CATALOG_SEARCH_PRODUCT, params=(product_id, ))
        return response[0]



"""Download Landsat catalog from Google and WRS2 Path/Row shapefile from USGS
and export them to a Spatialite-enabled SQLite database.
"""

from collections import OrderedDict
import csv
from datetime import datetime
from dateutil.parser import isoparse
import os
import shutil
import sqlite3
import tempfile

import fiona
from appdirs import user_data_dir
from shapely import wkt
from shapely.geometry import shape
from tqdm import tqdm

from pylandsat import queries
from pylandsat import utils


def _flatten(params):
    """Flatten the list objects in a list of DB-API parameters,
    e.g. `[1, [2, 3]]` becomes `[1, 2, 3]`.
    """
    params_flat = []
    for param in params:
        if isinstance(param, list):
            for item in param:
                params_flat.append(item)
        else:
            params_flat.append(param)
    return params_flat


def _format_placeholders(query, params):
    """Replace a single '?' DB-API placeholder after each 'IN'
    statement by an array of placeholders, e.g. `IN ?` becomes
    `IN (?, ?, ?)`. Also flatten the given parameter list.
    (This is because lists are not supported as DB-API parameters.)
    """
    n_placeholders = [len(p) for p in params if isinstance(p, list)]
    for n in n_placeholders:
        placeholders = ','.join(list('?' * n))
        query = query.replace('IN ?', 'IN ({})'.format(placeholders), 1)
    return query, _flatten(params)


class LandsatDB:
    """Initialize and query a Spatialite-enabled SQLite database."""

    def __init__(self):
        os.makedirs(user_data_dir('pylandsat'), exist_ok=True)
        self.path = os.path.join(user_data_dir('pylandsat'), 'landsat.db')

    def connect(self):
        """Connect to the DB and enable Spatialite."""
        conn = sqlite3.connect(self.path)
        conn.enable_load_extension(True)
        conn.execute("SELECT load_extension('mod_spatialite');")
        return conn

    def query(self, query, params=None):
        """Perform an SQL query and returns the result as a dict."""
        conn = self.connect()
        conn.row_factory = sqlite3.Row
        c = conn.cursor()
        if params:
            query, params = _format_placeholders(query, params)
            c.execute(query, params)
        else:
            c.execute(query)
        rows = c.fetchall()
        conn.close()
        return [dict(row) for row in rows]


def _parse_datestring(datestring):
    """Parse ISO-8601 date string from the index.csv file."""
    date = isoparse(datestring[:26])
    return int(date.timestamp())


def _parse_row(row):
    """Parse a row from the index.csv file."""
    INDEXES = [1, 0, 9, 10, 7, 11]
    product_id, scene_id, path, row, sensing_time, cloud_cover = (row[i] for i in INDEXES)
    path, row = int(path), int(row)
    cloud_cover = float(cloud_cover)
    sensing_time = _parse_datestring(sensing_time)
    return (product_id, scene_id, path, row, sensing_time, cloud_cover)


def sync_catalog():
    """Download Landsat catalog from Google and update the SQLite database
    accordingly.
    """
    CATALOG_URL = 'https://storage.googleapis.com/gcp-public-data-landsat/index.csv.gz'
    with tempfile.TemporaryDirectory(prefix="pylandsat_") as tmpdir:
        fpath = utils.download_file(CATALOG_URL, tmpdir, progressbar=True)
        fpath = utils.decompress(fpath, remove_archive=True)

        # Create database and 'catalog' table
        db = LandsatDB()
        conn = db.connect()
        c = conn.cursor()
        c.execute(queries.CATALOG_CREATE)
        conn.commit()

        # Insert CSV rows into the SQLite database
        length = sum(1 for line in open(fpath))
        progress = tqdm(total=length, unit=' rows')
        with open(fpath) as src:
            reader = csv.reader(src)
            _ = reader.__next__()  # ignore header
            for row in reader:
                if row[1]:
                    c.execute(queries.CATALOG_UPDATE, _parse_row(row))
                progress.update(1)
        progress.close()
        conn.commit()
        conn.close()


def sync_wrs():
    """Download WRS2 descending shapefile from USGS and export it to a
    Spatialite-enabled SQLite table.
    """
    WRS_URL = 'https://prd-wret.s3.us-west-2.amazonaws.com/assets/palladium/production/s3fs-public/atoms/files/WRS2_descending_0.zip'
    
    with tempfile.TemporaryDirectory(prefix="pylandsat_") as tmpdir:
        fpath = utils.download_file(WRS_URL, tmpdir, progressbar=True)

        # Connect to the database, init spatial metadata and create the table
        db = LandsatDB()
        conn = db.connect()
        c = conn.cursor()
        c.executescript(queries.WRS_CREATE)
        conn.commit()

        def _to_wkt(feature):
            geom = shape(feature['geometry'])
            return wkt.dumps(geom, rounding_precision=8)

        # Insert values and create the spatial index
        collection = fiona.open('/WRS2_descending.shp', vfs='zip://' + fpath)
        values = ((f['properties']['PATH'], f['properties']['ROW'], _to_wkt(f))
                for f in collection)
        c.executemany(queries.WRS_UPDATE, values)
        c.execute(queries.WRS_INDEX)

        conn.commit()
        conn.close()



"""Pre-made SQL queries."""

# Create catalog table
CATALOG_CREATE = """
CREATE TABLE IF NOT EXISTS catalog (
    product_id TEXT PRIMARY KEY NOT NULL,
    scene_id TEXT,
    path INTEGER,
    row INTEGER,
    sensing_time INTEGER,
    cloud_cover FLOAT
);"""

# Insert values into catalog table
CATALOG_UPDATE = """
INSERT OR IGNORE INTO catalog (product_id, scene_id, path, row,
  sensing_time, cloud_cover)
  VALUES (?, ?, ?, ?, ?, ?);"""

# Create index on path/row
CATALOG_INDEX = "CREATE INDEX idx_catalog_pathrow ON catalog (path, row);"

# Create the WRS table
WRS_CREATE = """
SELECT InitSpatialMetadata(1);
CREATE TABLE IF NOT EXISTS wrs (
    path INTEGER,
    row INTEGER,
    PRIMARY KEY (path, row)
);
SELECT AddGeometryColumn('wrs', 'geom', 4326, 'GEOMETRY', 'XY');
"""

# Insert values into the wrs table
WRS_UPDATE = """
INSERT OR IGNORE INTO wrs (path, row, geom)
  VALUES (?, ?, GeomFromText(?, 4326));
"""

# Create a spatial index on the wrs table
WRS_INDEX = "SELECT CreateSpatialIndex('wrs', 'geom');"

# Find path/row that intersect an input geometry
WRS_SEARCH = """
SELECT path, row, AsText(geom) AS geom FROM wrs
WHERE Intersects(geom, GeomFromText(?, 4326))
  AND wrs.ROWID IN (
    SELECT ROWID FROM SpatialIndex
    WHERE f_table_name = 'wrs'
    AND search_frame = GeomFromText(?, 4326)
  );
"""

# Search the catalog using path and row as spatial filtering
CATALOG_SEARCH_PATHROW = """
SELECT catalog.product_id, catalog.scene_id, catalog.path, catalog.row,
  catalog.sensing_time, catalog.cloud_cover, AsText(wrs.geom) AS geom
FROM catalog
INNER JOIN wrs ON wrs.path = catalog.path AND wrs.row = catalog.row
WHERE catalog.path IN ? AND catalog.row IN ?
  AND catalog.sensing_time BETWEEN ? AND ?
  AND catalog.cloud_cover <= ?
  AND SUBSTR(catalog.product_id, 1, 4) IN ?
  AND SUBSTR(catalog.product_id, -2, 2) IN ?
"""

# Search the catalog using an user-provided geometry as spatial filtering
# Tuple to provide: (geom, begin date, end date, max cloud cover, list of
# sensor ids, list of tiers, geom)
CATALOG_SEARCH_GEOM = """
SELECT catalog.product_id, catalog.scene_id, catalog.path, catalog.row,
  catalog.sensing_time, catalog.cloud_cover, AsText(wrs.geom) AS geom
FROM catalog
INNER JOIN wrs ON wrs.path = catalog.path AND wrs.row = catalog.row
WHERE Intersects(wrs.geom, GeomFromText(?, 4326))
  AND catalog.sensing_time BETWEEN ? AND ?
  AND catalog.cloud_cover <= ?
  AND SUBSTR(catalog.product_id, 1, 4) IN ?
  AND SUBSTR(catalog.product_id, -2, 2) IN ?
  AND wrs.ROWID IN (
    SELECT ROWID FROM SpatialIndex
    WHERE f_table_name = 'wrs'
    AND search_frame = GeomFromText(?, 4326)
  );
"""

# Get metadata for a given product ID
CATALOG_SEARCH_PRODUCT = """
SELECT catalog.product_id, catalog.scene_id, catalog.path, catalog.row,
  date(catalog.sensing_time, 'unixepoch') AS sensing_time,
  catalog.cloud_cover, AsText(wrs.geom) AS geom
FROM catalog
INNER JOIN wrs ON catalog.path = wrs.path
  AND catalog.row = wrs.row
WHERE
  product_id = ?
"""

# Count rows in a given table
COUNT = "SELECT COUNT(*) FROM ?"

# List all tables in the database
LIST_TABLES = "SELECT name FROM sqlite_master WHERE type='table'"
